# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus
from aerospike import exception as e
from aerospike import predicates as p
import time

import aerospike


def get_geo_object():
    geo_object = aerospike.GeoJSON(
        {
            "type": "Polygon",
            "coordinates": [
                [
                    [-124.500000, 37.000000],
                    [-125.000000, 37.000000],
                    [-121.000000, 38.080000],
                    [-122.500000, 38.080000],
                    [-124.500000, 37.000000],
                ]
            ],
        }
    )
    return geo_object


def add_geo_indexes(connection):
    try:
        connection.index_geo2dsphere_create("test", "demo", "loc", "loc_index")
    except (e.IndexFoundError):
        pass

    try:
        connection.index_geo2dsphere_create("test", "demo", "loc_polygon", "loc_polygon_index")
    except (e.IndexFoundError):
        pass

    try:
        connection.index_geo2dsphere_create("test", "demo", "loc_circle", "loc_circle_index")
    except (e.IndexFoundError):
        pass

    try:
        connection.index_list_create("test", "demo", "geo_list", aerospike.INDEX_GEO2DSPHERE, "geo_list_index")
    except (e.IndexFoundError):
        pass

    try:
        connection.index_map_keys_create(
            "test", "demo", "geo_map_keys", aerospike.INDEX_GEO2DSPHERE, "geo_map_key_index"
        )
    except (e.IndexFoundError):
        pass

    try:
        connection.index_map_values_create(
            "test", "demo", "geo_map_vals", aerospike.INDEX_GEO2DSPHERE, "geo_map_val_index"
        )
    except (e.IndexFoundError):
        pass

    try:
        connection.index_list_create("test", "demo", "geo_loc_list", aerospike.INDEX_GEO2DSPHERE, "geo_loc_list_index")
    except (e.IndexFoundError):
        pass

    try:
        connection.index_map_keys_create(
            "test", "demo", "geo_loc_mk", aerospike.INDEX_GEO2DSPHERE, "geo_loc_map_key_index"
        )
    except (e.IndexFoundError):
        pass

    try:
        connection.index_map_values_create(
            "test", "demo", "geo_loc_mv", aerospike.INDEX_GEO2DSPHERE, "geo_loc_map_val_index"
        )
    except (e.IndexFoundError):
        pass


def add_geo_data(connection):
    pre = '{"type": "Point", "coordinates"'
    suf = "]}"
    for i in range(10):
        lng = 1220 - (2 * i)
        lat = 375 + (2 * i)
        key = ("test", "demo", i)
        s = "{0}: [-{1}.{2}, {3}.{4}{5}".format(pre, (lng // 10), (lng % 10), (lat // 10), (lat % 10), suf)
        geo_object = aerospike.geojson(s)
        geo_list = [geo_object]
        geo_map_key = {geo_object: i}
        geo_map_val = {i: geo_object}
        connection.put(
            key, {"loc": geo_object, "geo_list": geo_list, "geo_map_keys": geo_map_key, "geo_map_vals": geo_map_val}
        )

    key = ("test", "demo", "polygon")
    geo_object_polygon = aerospike.GeoJSON(
        {
            "type": "Polygon",
            "coordinates": [
                [
                    [-122.500000, 37.000000],
                    [-121.000000, 37.000000],
                    [-121.000000, 38.080000],
                    [-122.500000, 38.080000],
                    [-122.500000, 37.000000],
                ]
            ],
        }
    )

    geo_loc_list = [geo_object_polygon]
    geo_loc_mk = {geo_object_polygon: 1}
    geo_loc_mv = {2: geo_object_polygon}
    connection.put(
        key,
        {
            "loc_polygon": geo_object_polygon,
            "geo_loc_list": geo_loc_list,
            "geo_loc_mk": geo_loc_mk,
            "geo_loc_mv": geo_loc_mv,
        },
    )

    key = ("test", "demo", "polygon2")
    geo_object_polygon = aerospike.GeoJSON(
        {
            "type": "Polygon",
            "coordinates": [
                [
                    [-52.500000, 37.000000],
                    [-51.000000, 37.000000],
                    [-51.000000, 38.080000],
                    [-52.500000, 38.080000],
                    [-52.500000, 37.000000],
                ]
            ],
        }
    )

    geo_loc_list = [geo_object_polygon]
    geo_loc_mk = {geo_object_polygon: 1}
    geo_loc_mv = {2: geo_object_polygon}
    connection.put(
        key,
        {
            "loc_polygon": geo_object_polygon,
            "geo_loc_list": geo_loc_list,
            "geo_loc_mk": geo_loc_mk,
            "geo_loc_mv": geo_loc_mv,
        },
    )


def remove_geo_indexes(connection):
    try:
        connection.index_remove("test", "loc_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "loc_polygon_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "loc_circle_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_list_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_map_key_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_map_val_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_loc_list_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_loc_map_key_index")
    except Exception:
        pass

    try:
        connection.index_remove("test", "geo_loc_map_val_index")
    except Exception:
        pass


def remove_geo_data(connection):
    for i in range(10):
        key = ("test", "demo", i)
        connection.remove(key)

    key = ("test", "demo", "polygon")
    connection.remove(key)

    key = ("test", "demo", "polygon2")
    connection.remove(key)


class TestGeospatial(object):
    def setup_class(cls):
        """
        Setup method.
        """
        cls.connection_setup_functions = (add_geo_indexes, add_geo_data)

        cls.connection_teardown_functions = (remove_geo_indexes, remove_geo_data)

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        as_connection = connection_with_config_funcs
        self.keys = []
        if not self.skip_old_server:
            key = ("test", "demo", "circle")
            geo_circle = aerospike.GeoJSON({"type": "AeroCircle", "coordinates": [[-122.0, 37.0], 250.2]})
            as_connection.put(key, {"loc_circle": geo_circle})
            self.keys.append(key)

        def teardown():
            for key in self.keys:
                as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_geospatial_put_get_positive(self):
        """
        Perform a get and put with multiple bins including geospatial bin
        """
        key = ("test", "demo", "single_geo_put")
        geo_object_single = aerospike.GeoJSON({"type": "Point", "coordinates": [42.34, 58.62]})
        geo_object_dict = aerospike.GeoJSON({"type": "Point", "coordinates": [56.34, 69.62]})

        self.as_connection.put(
            key,
            {
                "loc": geo_object_single,
                "int_bin": 2,
                "string_bin": "str",
                "dict_bin": {"a": 1, "b": 2, "geo": geo_object_dict},
            },
        )

        key, _, bins = self.as_connection.get(key)

        expected = {
            "loc": {"coordinates": [42.34, 58.62], "type": "Point"},
            "int_bin": 2,
            "string_bin": "str",
            "dict_bin": {"a": 1, "b": 2, "geo": {"coordinates": [56.34, 69.62], "type": "Point"}},
        }
        for b in bins:
            assert b in expected

        self.as_connection.remove(key)

    def test_geospatial_positive_query(self):
        """
        Perform a positive geospatial query for a polygon
        """
        records = []
        query = self.as_connection.query("test", "demo")
        geo_object2 = aerospike.GeoJSON(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3
        expected = [
            {"coordinates": [-122.0, 37.5], "type": "Point"},
            {"coordinates": [-121.8, 37.7], "type": "Point"},
            {"coordinates": [-121.6, 37.9], "type": "Point"},
        ]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_positive_query_outside_shape(self):
        """
        Perform a positive geospatial query for polygon where all points
        are outside polygon
        """
        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-126.500000, 37.000000],
                        [-124.000000, 37.000000],
                        [-124.000000, 38.080000],
                        [-126.500000, 38.080000],
                        [-126.500000, 37.000000],
                    ]
                ],
            }
        )

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 0

    def test_geospatial_positive_query_without_set(self):
        """
        Perform a positive geospatial query for a polygon without a set
        """
        keys = []
        pre = '{"type": "Point", "coordinates"'
        suf = "]}"
        for i in range(1, 10):
            lng = 1220 - (2 * i)
            lat = 375 + (2 * i)
            key = ("test", None, i)
            s = "{0}: [-{1}.{2}, {3}.{4}{5}".format(pre, (lng // 10), (lng % 10), (lat // 10), (lat % 10), suf)
            geo_object = aerospike.geojson(s)
            self.as_connection.put(key, {"loc": geo_object})
            keys.append(key)

        try:
            self.as_connection.index_geo2dsphere_create("test", None, "loc", "loc_index_no_set")
        except (e.IndexFoundError):
            pass

        records = []
        query = self.as_connection.query("test", None)

        geo_object2 = aerospike.GeoJSON(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)
        try:
            self.as_connection.index_remove("test", "loc_index_no_set")
        except (Exception):
            pass

        for key in keys:
            self.as_connection.remove(key)

        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            assert len(records) == 2
        else:
            assert len(records) == 5

        expected = [{"coordinates": [-121.8, 37.7], "type": "Point"}, {"coordinates": [-121.6, 37.9], "type": "Point"}]

        for r in records:
            if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
                assert r["loc"].unwrap() in expected
            else:
                expected = [
                    {"coordinates": [-122.0, 37.5], "type": "Point"},
                    {"coordinates": [-121.8, 37.7], "type": "Point"},
                    {"coordinates": [-121.6, 37.9], "type": "Point"},
                ]
                assert r["loc"].unwrap() in expected

    def test_geospatial_positive_query_for_circle(self):
        """
        Perform a positive geospatial query for a circle
        """
        if TestGeospatial.skip_old_server is True:
            pytest.skip("Server does not support apply on AeroCircle for GeoJSON")

        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "AeroCircle", "coordinates": [[-122.0, 37.5], 250.2]})

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [{"coordinates": [-122.0, 37.5], "type": "Point"}]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geo_query_with_geo_within_radius_predicate(self):
        """
        Perform a positive geospatial query for a circle with helper
        """
        if TestGeospatial.skip_old_server is True:
            pytest.skip("Server does not support apply on AeroCircle for GeoJSON")

        records = []
        query = self.as_connection.query("test", "demo")

        query.where(p.geo_within_radius("loc", -122.0, 37.5, 250.2))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [{"coordinates": [-122.0, 37.5], "type": "Point"}]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_operate_positive(self):
        """
        Perform an operate operation with geospatial bin
        """

        geo_object_operate = aerospike.GeoJSON({"type": "Point", "coordinates": [43.45, 56.75]})
        key = ("test", "demo", "single_geo_operate")
        llist = [
            {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": {"no": geo_object_operate}},
            {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
        ]

        key, _, bins = self.as_connection.operate(key, llist)
        self.keys.append(key)
        assert type(bins["write_bin"]["no"]) == aerospike.GeoJSON
        assert bins["write_bin"]["no"].unwrap() == {"coordinates": [43.45, 56.75], "type": "Point"}

    def test_geospatial_wrap_positive(self):
        """
        Perform a positive wrap on geospatial data
        """
        geo_object = aerospike.GeoJSON(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-124.500000, 37.000000],
                        [-125.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-124.500000, 37.000000],
                    ]
                ],
            }
        )
        geo_object.wrap(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        assert geo_object.unwrap() == {
            "coordinates": [[[-122.5, 37.0], [-121.0, 37.0], [-121.0, 38.08], [-122.5, 38.08], [-122.5, 37.0]]],
            "type": "Polygon",
        }

    def test_geospatial_wrap_positive_with_query(self):
        """
        Perform a positive wrap on geospatial data followed by a query
        """
        geo_object_wrap = get_geo_object()

        geo_object_wrap.wrap(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        assert geo_object_wrap.unwrap() == {
            "coordinates": [[[-122.5, 37.0], [-121.0, 37.0], [-121.0, 38.08], [-122.5, 38.08], [-122.5, 37.0]]],
            "type": "Polygon",
        }

        records = []
        query = self.as_connection.query("test", "demo")
        query.where(p.geo_within_geojson_region("loc", geo_object_wrap.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3
        expected = [
            {"coordinates": [-122.0, 37.5], "type": "Point"},
            {"coordinates": [-121.8, 37.7], "type": "Point"},
            {"coordinates": [-121.6, 37.9], "type": "Point"},
        ]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_loads_positive(self):
        """
        Perform a positive loads on geoJSON raw string
        """
        geo_object = get_geo_object()
        geo_object.loads(
            '{"type": "Polygon", "coordinates": [[[-122.500000, 37.000000],\
             [-121.000000, 37.000000], [-121.000000, 38.080000],\
             [-122.500000, 38.080000], [-122.500000, 37.000000]]]}'
        )

        assert geo_object.unwrap() == {
            "coordinates": [[[-122.5, 37.0], [-121.0, 37.0], [-121.0, 38.08], [-122.5, 38.08], [-122.5, 37.0]]],
            "type": "Polygon",
        }

    def test_geospatial_loads_positive_with_query(self):
        """
        Perform a positive loads on geoJSON raw string followed by a query
        """
        geo_object_loads = get_geo_object()

        geo_object_loads.loads(
            '{"type": "Polygon", "coordinates": [[[-122.500000, 37.000000],\
             [-121.000000, 37.000000], [-121.000000, 38.080000],\
             [-122.500000, 38.080000], [-122.500000, 37.000000]]]}'
        )

        assert geo_object_loads.unwrap() == {
            "coordinates": [[[-122.5, 37.0], [-121.0, 37.0], [-121.0, 38.08], [-122.5, 38.08], [-122.5, 37.0]]],
            "type": "Polygon",
        }

        records = []
        query = self.as_connection.query("test", "demo")
        query.where(p.geo_within_geojson_region("loc", geo_object_loads.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3
        expected = [
            {"coordinates": [-122.0, 37.5], "type": "Point"},
            {"coordinates": [-121.8, 37.7], "type": "Point"},
            {"coordinates": [-121.6, 37.9], "type": "Point"},
        ]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_dumps_positive(self):
        """
        Perform a positive dumps. Verify using str
        """
        geo_object = get_geo_object()

        geojson_str = geo_object.dumps()
        assert isinstance(geojson_str, str)

        obj = aerospike.geojson(geojson_str)
        assert obj.unwrap() == geo_object.unwrap()

    def test_geojson_str(self):
        """
        verify that str representation of geojson object is correct
        """
        geo_object = get_geo_object()

        geojson_str = str(geo_object)
        assert isinstance(geojson_str, str)

        obj = aerospike.geojson(geojson_str)
        assert obj.unwrap() == geo_object.unwrap()

    def test_geospatial_repr_positive(self):
        """
        Perform a positive repr. Verify using eval()
        """

        geo_object = get_geo_object()

        geojson_str = eval(repr(geo_object))
        assert isinstance(geojson_str, str)

        obj = aerospike.geojson(geojson_str)
        assert obj.unwrap() == geo_object.unwrap()

    def test_geospatial_put_get_positive_with_geodata(self):
        """
        Perform a get and put with multiple bins including geospatial bin
        using geodata method
        """

        key = ("test", "demo", "single_geo_put")
        geo_object_single = aerospike.geodata({"type": "Point", "coordinates": [42.34, 58.62]})
        geo_object_dict = aerospike.geodata({"type": "Point", "coordinates": [56.34, 69.62]})

        self.as_connection.put(
            key,
            {
                "loc": geo_object_single,
                "int_bin": 2,
                "string_bin": "str",
                "dict_bin": {"a": 1, "b": 2, "geo": geo_object_dict},
            },
        )

        key, _, bins = self.as_connection.get(key)

        expected = {
            "loc": {"coordinates": [42.34, 58.62], "type": "Point"},
            "int_bin": 2,
            "string_bin": "str",
            "dict_bin": {"a": 1, "b": 2, "geo": {"coordinates": [56.34, 69.62], "type": "Point"}},
        }
        for b in bins:
            assert b in expected

        self.as_connection.remove(key)

    def test_geospatial_put_get_positive_with_geojson(self):
        """
        Perform a get and put with multiple bins including geospatial bin
        using geodata method
        """

        key = ("test", "demo", "single_geo_put")
        geo_object_single = aerospike.geojson('{"type": "Point", "coordinates": [42.34, 58.62] }')
        geo_object_dict = aerospike.geojson('{"type": "Point", "coordinates": [56.34, 69.62] }')

        self.as_connection.put(
            key,
            {
                "loc": geo_object_single,
                "int_bin": 2,
                "string_bin": "str",
                "dict_bin": {"a": 1, "b": 2, "geo": geo_object_dict},
            },
        )

        key, _, bins = self.as_connection.get(key)

        expected = {
            "loc": {"coordinates": [42.34, 58.62], "type": "Point"},
            "int_bin": 2,
            "string_bin": "str",
            "dict_bin": {"a": 1, "b": 2, "geo": {"coordinates": [56.34, 69.62], "type": "Point"}},
        }
        for b in bins:
            assert b in expected

        self.as_connection.remove(key)

    def test_geospatial_positive_query_with_geodata(self):
        """
        Perform a positive geospatial query for a polygon with geodata
        """
        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.geodata(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3
        expected = [
            {"coordinates": [-122.0, 37.5], "type": "Point"},
            {"coordinates": [-121.8, 37.7], "type": "Point"},
            {"coordinates": [-121.6, 37.9], "type": "Point"},
        ]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_positive_query_with_geojson(self):
        """
        Perform a positive geospatial query for a polygon with geojson
        """
        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.geojson(
            '{"type": "Polygon", "coordinates": [[[-122.500000, 37.000000], \
            [-121.000000, 37.000000], [-121.000000, 38.080000],\
            [-122.500000, 38.080000], [-122.500000, 37.000000]]]}'
        )

        query.where(p.geo_within_geojson_region("loc", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3
        expected = [
            {"coordinates": [-122.0, 37.5], "type": "Point"},
            {"coordinates": [-121.8, 37.7], "type": "Point"},
            {"coordinates": [-121.6, 37.9], "type": "Point"},
        ]
        for r in records:
            assert r["loc"].unwrap() in expected

    def test_geospatial_2dindex_positive(self):
        """
        Perform a positive 2d index creation
        """
        try:
            status = self.as_connection.index_remove("test", "loc_index")
            time.sleep(2)
        except Exception:
            pass

        status = self.as_connection.index_geo2dsphere_create("test", "demo", "loc", "loc_index")

        assert status == 0

    def test_geospatial_2dindex_positive_with_policy(self):
        """
        Perform a positive 2d index creation with policy
        """
        try:
            status = self.as_connection.index_remove("test", "loc_index")
            time.sleep(2)
        except Exception:
            pass

        status = self.as_connection.index_geo2dsphere_create("test", "demo", "loc", "loc_index", {"timeout": 2000})

        assert status == 0

    def test_geospatial_positive_query_with_point(self):
        """
        Perform a positive geospatial query for a point
        """
        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-121.700000, 37.200000]})

        query.where(p.geo_contains_geojson_point("loc_polygon", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [
            {
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
                "type": "Polygon",
            }
        ]

        for r in records:
            assert r["loc_polygon"].unwrap() in expected

    def test_geospatial_positive_query_with_point_outside_polygon(self):
        """
        Perform a positive geospatial query for a point outside polygon
        """
        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-123.700000, 37.200000]})

        query.where(p.geo_contains_geojson_point("loc_polygon", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 0

    def test_geospatial_positive_query_with_point_in_aerocircle(self):
        """
        Perform a positive geospatial query for a point in aerocircle
        """
        if TestGeospatial.skip_old_server is True:
            pytest.skip("Server does not support apply on AeroCircle for GeoJSON")

        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-122.000000, 37.000000]})

        query.where(p.geo_contains_geojson_point("loc_circle", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [{"coordinates": [[-122.0, 37.0], 250.2], "type": "AeroCircle"}]
        for r in records:
            assert r["loc_circle"].unwrap() in expected

    def test_geospatial_positive_query_with_point_in_aerocircle_int(self):
        """
        Perform a positive geospatial query for a point in aerocircle
        """
        if TestGeospatial.skip_old_server is True:
            pytest.skip("Server does not support apply on AeroCircle for GeoJSON")

        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-122, 37]})

        query.where(p.geo_contains_geojson_point("loc_circle", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [{"coordinates": [[-122.0, 37.0], 250.2], "type": "AeroCircle"}]
        for r in records:
            assert r["loc_circle"].unwrap() in expected

    def test_geospatial_positive_query_with_point_outside_aerocircle(self):
        """
        Perform a positive geospatial query for a point in aerocircle
        """
        if TestGeospatial.skip_old_server is True:
            pytest.skip("Server does not support apply on AeroCircle for GeoJSON")

        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-122.0, 48.0]})

        query.where(p.geo_contains_geojson_point("loc_circle", geo_object2.dumps()))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 0

    def test_geospatial_positive_query_with_point_helper_method(self):
        """
        Perform a positive geospatial query for a point with helper method
        """
        records = []
        query = self.as_connection.query("test", "demo")

        query.where(p.geo_contains_point("loc_polygon", -121.7, 37.2))

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 1
        expected = [
            {
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
                "type": "Polygon",
            }
        ]

        for r in records:
            assert r["loc_polygon"].unwrap() in expected

    @pytest.mark.parametrize(
        "bin_name, idx_type",
        (
            ("geo_list", aerospike.INDEX_TYPE_LIST),
            ("geo_map_keys", aerospike.INDEX_TYPE_MAPKEYS),
            ("geo_map_vals", aerospike.INDEX_TYPE_MAPVALUES),
        ),
    )
    def test_geospatial_within_radius_pred(self, bin_name, idx_type):

        records = []
        query = self.as_connection.query("test", "demo")

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        predicate = p.geo_within_radius(bin_name, -122.0, 37.5, 250.2, idx_type)

        query.where(predicate)
        query.foreach(callback)

        assert len(records) == 1

    @pytest.mark.parametrize(
        "bin_name, idx_type",
        (
            ("geo_list", aerospike.INDEX_TYPE_LIST),
            ("geo_map_keys", aerospike.INDEX_TYPE_MAPKEYS),
            ("geo_map_vals", aerospike.INDEX_TYPE_MAPVALUES),
        ),
    )
    def test_geospatial_within_geojson_region_pred(self, bin_name, idx_type):

        records = []
        query = self.as_connection.query("test", "demo")

        geo_object2 = aerospike.geodata(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-122.500000, 37.000000],
                        [-121.000000, 37.000000],
                        [-121.000000, 38.080000],
                        [-122.500000, 38.080000],
                        [-122.500000, 37.000000],
                    ]
                ],
            }
        )

        predicate = p.geo_within_geojson_region(bin_name, geo_object2.dumps(), idx_type)

        query.where(predicate)

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.foreach(callback)

        assert len(records) == 3

    def test_store_multipolygon(self):

        polygons = [
            [
                [
                    [-124.500000, 37.000000],
                    [-125.000000, 37.000000],
                    [-121.000000, 38.080000],
                    [-122.500000, 38.080000],
                    [-124.500000, 37.000000],
                ]
            ],
            [
                [
                    [-24.500000, 37.000000],
                    [-25.000000, 37.000000],
                    [-21.000000, 38.080000],
                    [-22.500000, 38.080000],
                    [-24.500000, 37.000000],
                ]
            ],
        ]
        geo_object = aerospike.GeoJSON({"type": "MultiPolygon", "coordinates": polygons})

        key = ("test", "demo", "multipoly")
        self.as_connection.put(key, {"multi": geo_object})

        _, _, bins = self.as_connection.get(key)

        geo_returned = bins["multi"].unwrap()
        assert geo_returned["type"] == "MultiPolygon"
        assert geo_returned["coordinates"] == polygons

        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "bin_name, idx_type",
        (
            ("geo_loc_list", aerospike.INDEX_TYPE_LIST),
            ("geo_loc_mk", aerospike.INDEX_TYPE_MAPKEYS),
            ("geo_loc_mv", aerospike.INDEX_TYPE_MAPVALUES),
        ),
    )
    def test_geospatial_contains_point_pred(self, bin_name, idx_type):

        records = []
        query = self.as_connection.query("test", "demo")
        lat = -122.45
        lon = 37.5

        predicate = p.geo_contains_point(bin_name, lat, lon, idx_type)

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.where(predicate)
        query.foreach(callback)

        assert len(records) == 1

    @pytest.mark.parametrize(
        "bin_name, idx_type",
        (
            ("geo_loc_list", aerospike.INDEX_TYPE_LIST),
            ("geo_loc_mk", aerospike.INDEX_TYPE_MAPKEYS),
            ("geo_loc_mv", aerospike.INDEX_TYPE_MAPVALUES),
        ),
    )
    def test_geospatial_contains_json_point_pred(self, bin_name, idx_type):

        records = []
        query = self.as_connection.query("test", "demo")
        lat = -122.45
        lon = 37.5
        point_list = [lat, lon]

        point = aerospike.GeoJSON({"type": "Point", "coordinates": point_list})

        predicate = p.geo_contains_geojson_point(bin_name, point.dumps(), idx_type)

        def callback(input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query.where(predicate)
        query.foreach(callback)

        assert len(records) == 1

    def test_geospatial_object_not_dict_or_string(self):
        """
        The geospatial object is not a dictionary or string
        """
        with pytest.raises(e.ParamError) as err_info:
            aerospike.GeoJSON(1)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_geospatial_object_non_json_serializable_string(self):
        """
        The geospatial object is not a json serializable string
        """
        with pytest.raises(e.ClientError) as err_info:
            aerospike.GeoJSON("abc")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_geospatial_object_wrap_non_dict(self):
        """
        The geospatial object provided to wrap() is not a dictionary
        """
        geo_object = get_geo_object()
        with pytest.raises(e.ParamError) as err_info:
            geo_object.wrap("abc")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_geospatial_object_loads_non_dict(self):
        """
        The geospatial object provided to loads() is not a dictionary
        """
        geo_object = get_geo_object()
        with pytest.raises(e.ClientError) as err_info:
            geo_object.loads('{"abc"}')

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_geospatial_2dindex_set_length_extra(self):
        """
        Perform a 2d creation with set length exceeding limit
        """
        set_name = "a" * 100

        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_geo2dsphere_create("test", set_name, "loc", "loc_index_creation_should_fail")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    @pytest.mark.skip(reason="These raise system errors")
    @pytest.mark.parametrize(
        "method", ["geo_within_geojson_region", "geo_contains_geojson_point", "geo_within_radius", "geo_contains_point"]
    )
    def test_call_geo_predicates_with_wrong_args(self, method):
        query = self.as_connection.query("test", "demo")
        predicate = getattr(p, method)

        with pytest.raises(e.ParamError) as err_info:
            query.where(predicate())

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


class TestClose:
    def setup_class(cls):
        config = TestBaseClass.get_connection_config()
        TestClose.hostlist = config["hosts"]
        TestClose.user = config["user"]
        TestClose.password = config["password"]
        TestClose.auth_mode = config["policies"]["auth_mode"]

    def test_pos_close(self):
        """
        Invoke close() after positive connect
        """
        self.client = TestBaseClass.get_new_connection()
        self.closeobject = self.client.close()
        assert self.closeobject is None

    def test_pos_close_without_connection(self):
        """
        Invoke close() without connection
        """
        config = TestBaseClass.get_connection_config()
        self.client = aerospike.client(config)
        self.client.close()

        try:
            self.closeobject = self.client.close()

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_close(self):
        """
        Invoke close() after negative connect
        """
        config = {"hosts": [("127.0.0.1", 2000)]}

        with pytest.raises(Exception):
            self.client = aerospike.client(config).connect()
        with pytest.raises(AttributeError) as attributeError:
            self.closeobject = self.client.close()
        assert "has no attribute" in str(attributeError.value)

    def test_close_twice_in_a_row(self):
        """
        Client call itself establishes connection.
        """
        config = TestBaseClass.get_connection_config()
        if TestClose.user is None and TestClose.password is None:
            self.client = aerospike.client(config).connect()
        else:
            self.client = aerospike.client(config).connect(TestClose.user, TestClose.password)

        assert self.client.is_connected()
        self.client.close()
        assert self.client.is_connected() is False

        # This second call should not raise any errors
        self.client.close()
        assert self.client.is_connected() is False



# Get nonexistent record
try:
    client.get(keyTuple)
except ex.RecordNotFound as e:
    print("Error: {0} [{1}]".format(e.msg, e.code))
    # Error: 127.0.0.1:3000 AEROSPIKE_ERR_RECORD_NOT_FOUND [2]

# Get existing record
client.put(keyTuple, {'bin1': 4})
(key, meta, bins) = client.get(keyTuple)

print(key) # ('test', 'demo', None, bytearray(b'...'))
print(meta) # {'ttl': 2592000, 'gen': 1}
print(bins) # {'bin1': 4}



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestRevokePrivilege(TestBaseClass):

    config = TestBaseClass.get_connection_config()

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = self.config
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except e.InvalidRole:
            pass
        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        )
        time.sleep(2)
        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except e.InvalidRole:
            pass
        self.client.close()

    def test_admin_revoke_privileges_no_parameters(self):
        """
        Revoke privileges with no parameters
        """
        with pytest.raises(TypeError):
            self.client.admin_revoke_privileges()

    @pytest.mark.parametrize(
        "privs",
        [
            ([{"code": aerospike.PRIV_DATA_ADMIN, "ns": "", "set": ""}]),
            ([{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}]),
            ([{"code": aerospike.PRIV_WRITE, "ns": "test", "set": "demo"}]),
            ([{"code": aerospike.PRIV_READ_WRITE, "ns": "test", "set": "demo"}]),
            ([{"code": aerospike.PRIV_READ_WRITE_UDF, "ns": "test", "set": "demo"}]),
            ([{"code": aerospike.PRIV_TRUNCATE, "ns": "test", "set": "demo"}]),
            ([{"code": aerospike.PRIV_UDF_ADMIN, "ns": "", "set": ""}]),
            ([{"code": aerospike.PRIV_SINDEX_ADMIN, "ns": "", "set": ""}]),
        ],
    )
    def test_admin_revoke_privileges_all_positive(self, privs):
        """
        revoke privileges positive
        """
        status = self.client.admin_grant_privileges("usr-sys-admin-test", privs)

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}, *privs]

        status = self.client.admin_revoke_privileges("usr-sys-admin-test", privs)

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_admin_revoke_privileges_positive(self):
        """
        revoke privileges positive
        """
        status = self.client.admin_grant_privileges("usr-sys-admin-test", [{"code": aerospike.PRIV_READ}])

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [
            {"code": 0, "ns": "", "set": ""},
            {"code": 1, "ns": "", "set": ""},
            {"code": 10, "ns": "", "set": ""},
        ]

        status = self.client.admin_revoke_privileges("usr-sys-admin-test", [{"code": aerospike.PRIV_READ}])

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_admin_revoke_privileges_positive_write(self):
        """
        revoke write privileges positive
        """
        status = self.client.admin_grant_privileges("usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE}])

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [
            {"code": 0, "ns": "", "set": ""},
            {"code": 1, "ns": "", "set": ""},
            {"code": 13, "ns": "", "set": ""},
        ]

        status = self.client.admin_revoke_privileges("usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE}])

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_admin_revoke_privileges_positive_with_policy(self):
        """
        Revoke privileges positive with policy
        """
        status = self.client.admin_grant_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ}], {"timeout": 1000}
        )

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [
            {"code": 0, "ns": "", "set": ""},
            {"code": 1, "ns": "", "set": ""},
            {"code": 10, "ns": "", "set": ""},
        ]

        status = self.client.admin_revoke_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ}], {"timeout": 1000}
        )

        time.sleep(1)
        assert status == 0
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_admin_revoke_privileges_positive_with_policy_write(self):
        """
        Revoke write privileges positive with policy
        """
        status = self.client.admin_grant_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE}], {"timeout": 1000}
        )

        assert status == 0
        time.sleep(2)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [
            {"code": 0, "ns": "", "set": ""},
            {"code": 1, "ns": "", "set": ""},
            {"code": 13, "ns": "", "set": ""},
        ]

        status = self.client.admin_revoke_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE}], {"timeout": 1000}
        )

        time.sleep(1)
        assert status == 0
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_admin_revoke_privileges_positive_with_ns_set(self):
        """
        Revoke privileges positive with ns and set
        """
        status = self.client.admin_grant_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}]
        )

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [
            {"code": 0, "ns": "", "set": ""},
            {"code": 1, "ns": "", "set": ""},
            {"code": 10, "ns": "test", "set": "demo"},
        ]

        status = self.client.admin_revoke_privileges(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}]
        )

        time.sleep(1)
        assert status == 0
        roles = self.client.admin_query_role("usr-sys-admin-test")
        assert roles == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

    def test_revoke_privileges_incorrect_role_type(self):
        """
        role name not string
        """
        try:
            self.client.admin_revoke_privileges(1, [{"code": aerospike.PRIV_USER_ADMIN}])

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Role name should be a string"

    def test_revoke_privileges_unknown_privilege_type(self):
        """
        privilege type unknown
        """
        try:
            self.client.admin_revoke_privileges("usr-sys-admin-test", [{"code": 64}])
        except e.InvalidPrivilege as exception:
            assert exception.code == 72

    def test_revoke_privileges_incorrect_privilege_type(self):
        """
        privilege type incorrect
        """
        try:
            self.client.admin_revoke_privileges("usr-sys-admin-test", None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Privileges should be a list"

    def test_revoke_privileges_empty_list_privileges(self):
        """
        privilege type is an empty list
        """
        try:
            self.client.admin_revoke_privileges("usr-sys-admin-test", [])

        except e.InvalidPrivilege as exception:
            assert exception.code == 72
            assert exception.msg == "AEROSPIKE_INVALID_PRIVILEGE"



# -*- coding: utf-8 -*-

import pytest

from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("as_connection", "connection_config")
class TestInfo(object):
    def test_info_all(self):
        request = "statistics"
        nodes_info = self.as_connection.info_all(request)

        assert nodes_info is not None

        assert type(nodes_info) == dict

    def test_info_all_with_None_policy(self):
        request = "statistics"
        nodes_info = self.as_connection.info_all(request, None)

        assert nodes_info is not None

        assert type(nodes_info) == dict

    @pytest.mark.parametrize(
        "container_type, container_name",
        [("namespaces", "test"), ("sets", "demo"), ("bins", "names")],
        ids=("namespace", "sets", "bins"),
    )
    def test_positive_info_all(self, container_type, container_name):
        """
        Test to see whether a namespace, set,
        and bin exist after a key is added
        """
        key = ("test", "demo", "list_key")
        rec = {"names": ["John", "Marlen", "Steve"]}

        self.as_connection.put(key, rec)

        response = self.as_connection.info_all(container_type)
        self.as_connection.remove(key)
        found = False
        for keys in response.keys():
            for value in response[keys]:
                if value is not None:
                    if container_name in value:
                        found = True

        assert found

    def test_info_all_with_config_for_statistics_and_policy(self):

        request = "statistics"
        policy = {"timeout": 1000}
        # hosts = [host for host in self.connection_config["hosts"]]

        nodes_info = self.as_connection.info_all(request, policy)

        assert nodes_info is not None
        assert isinstance(nodes_info, dict)

    def test_info_all_for_invalid_request(self):

        request = "fake_request_string_not_real"
        # hosts = [host for host in self.connection_config["hosts"]]
        nodes_info = self.as_connection.info_all(request)

        assert isinstance(nodes_info, dict)
        assert nodes_info.values() is not None

    def test_info_all_with_none_request(self):
        """
        Test that sending None as the request raises an error
        """

        with pytest.raises(e.ParamError):
            self.as_connection.info_all(None)

    def test_info_all_without_parameters(self):

        with pytest.raises(TypeError):
            self.as_connection.info_all()

    def test_info_all_without_connection(self):
        """
        Test info positive for sets without connection
        """
        client1 = aerospike.client(self.connection_config)
        client1.close()

        with pytest.raises(e.ClusterError):
            client1.info_all("sets")

    def test_info_all_with_invalid_policy_type(self):
        """
        Test that sending a non dict/None as policy raises an error
        """

        with pytest.raises(e.ParamError):
            self.as_connection.info_all(None, [])



# Keys
keyTuples = [
    ('test', 'demo', '1'),
    ('test', 'demo', '2'),
    ('test', 'demo', '3'),
]
# Only insert two records with the first and second key
client.put(keyTuples[0], {'bin1': 'value'})
client.put(keyTuples[1], {'bin1': 'value'})

# Try to get records with all three keys
records = client.get_many(keyTuples)
# The third record tuple should have 'meta' and 'bins' set to none
# Because there is no third record
print(records[0])
print(records[1])
print(records[2])

# Expected output:
# (('test', 'demo', '1', bytearray(...)), {'ttl': 2592000, 'gen': 1}, {'bin1': 'value'})
# (('test', 'demo', '2', bytearray(...)), {'ttl': 2592000, 'gen': 1}, {'bin1': 'value'})
# (('test', 'demo', '3', bytearray(...)), None, None)
